/*================================================================
  Xsnake.c -> Version XWindow del juego SNAKE (clsico de los
              primeros videojuegos arcade). El juego consiste
              en mover una serpiente por la pantalla sin chocarse
              con las paredes ni con la cola, y recogiendo frutas
              que aumentaran el tamao de esta.
  (c) 1999 Santiago Romero Iglesias y Antonio Fernandez Fandos
 ================================================================*/

/*--- Ficheros de cabecera utilizados -------------------*/
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdio.h>
#include <X11/extensions/shape.h>
#include <Imlib.h>
#include <sys/time.h>
#include <unistd.h>

#include "xsnake.h"



/*=== Funcion principal main() ==========================*/
int main()
{
  Display *display;
  Window window, rootwin;
  int pant, anchura, altura;
  Pixmap pixmap, mask;
  GC gc;
  ImlibData *id;
  ImlibImage *imagen;
  XSetWindowAttributes attr;
  XEvent evento;
  Font fuente;
  char key;
  struct serpiente serpiente1;
  XSizeHints myhints;

  display = XOpenDisplay( NULL );
  XGetGeometry( display, XDefaultRootWindow(display), &rootwin,
                &pant, &pant, &anchura, &altura, &pant, &pant );

  id = Imlib_init(display);
  window = XCreateWindow(display,
           DefaultRootWindow(display),
           (anchura/2)-((ANCHURA_BLOQUE*(xmax+1))/2),
           (altura/2)-((ALTURA_BLOQUE*(ymax+1)+40)/2),
           ANCHURA_BLOQUE*(xmax+1), ALTURA_BLOQUE*(ymax+1)+40, 0,
           id->x.depth, InputOutput, id->x.visual, 0, &attr);

  myhints.min_width  =
  myhints.max_width  = ANCHURA_BLOQUE*(xmax+1) ;
  myhints.min_height =
  myhints.max_height = ALTURA_BLOQUE*(ymax+1)+40 ;
  myhints.x = (anchura/2)-((ANCHURA_BLOQUE*(xmax+1))/2);
  myhints.y = (altura/2)-((ALTURA_BLOQUE*(ymax+1)+40)/2);
  myhints.flags = PMinSize | PMaxSize | PPosition;
  XSetNormalHints( display, window, &myhints );

  pant = DefaultScreen(display);
  XSelectInput( display, window, KeyPressMask | ExposureMask  );
  XStoreName(display, window, "X-Snake");
  pixmap = XCreatePixmap( display, window,
              ANCHURA_BLOQUE*(xmax+1), ALTURA_BLOQUE*(ymax+1)+40,
              DefaultDepth(display, pant));

  if( pixmap == (Pixmap) NULL )
  {
    printf("\nError creando Pixmap.\n");
    printf("Saliendo de X-Snake...\n");
    XCloseDisplay(display);
    exit(1);
  }

  gc = XCreateGC( display, pixmap, 0, NULL );
  XSetForeground(display, gc, 0 );
  XFillRectangle( display, pixmap, gc, 0, 0,
                  ANCHURA_BLOQUE*(xmax+1), ALTURA_BLOQUE*(ymax+1)+40 );

  negro = ColorPorNombre( display, "black");
  amarillo = ColorPorNombre(display, "yellow");
  verde = ColorPorNombre( display, "green");
  fuente = XLoadFont( display, "9x15" );
  XSetFont( display, gc, fuente );

  /* cargamos las imagenes JPG para el juego en pixmaps */
  imagen=Imlib_load_image(id, "xsnake0.jpg");
  Imlib_render( id, imagen, ANCHURA_BLOQUE, ALTURA_BLOQUE );
  sprites[0] = (Pixmap) Imlib_move_image( id, imagen );
  imagen=Imlib_load_image(id, "xsnake1.jpg");
  Imlib_render( id, imagen, ANCHURA_BLOQUE, ALTURA_BLOQUE );
  sprites[1] = (Pixmap) Imlib_move_image( id, imagen );
  imagen=Imlib_load_image(id, "xsnake2.jpg");
  Imlib_render( id, imagen, ANCHURA_BLOQUE, ALTURA_BLOQUE );
  sprites[2] = (Pixmap) Imlib_move_image( id, imagen );
  imagen=Imlib_load_image(id, "xsnake3.jpg");
  Imlib_render( id, imagen, ANCHURA_BLOQUE*2, ALTURA_BLOQUE*2 );
  sprites[3] = (Pixmap) Imlib_move_image( id, imagen );

  srand( time(NULL) );
  DibujaTablero( display, pixmap, window, gc );
  InicializaMatrices();
  InicializaSerpiente( &serpiente1 );
  DibujaSerpiente(display, pixmap, window, gc, &serpiente1 );
  ActualizaPuntos(display, pixmap, window, gc, &serpiente1 );
  XMapWindow( display, window );

  while ( !done )
  {
    /* coger el tiempo actual */
    gettimeofday(&start, NULL);

   /* a continuacion miramos si hay algun evento en la
    cola de eventos. Si lo hay, lo recogemos y lo
    interpretamos. Si no lo hay, se repite el bucle. */
    while (QLength(display) > 0)
    {
      XNextEvent(display, &evento);
      if (evento.xany.window != window) continue;

      switch(evento.type)
      {
       case Expose:
           RedibujaPantalla(display, window, gc, pixmap,
                   evento.xexpose.x, evento.xexpose.y,
                   evento.xexpose.width, evento.xexpose.height );
           XFlush(display);
           break;

       case KeyPress:
           key = XKeycodeToKeysym(display, evento.xkey.keycode, 0);
           if( evento.xkey.keycode == ESC_KEY )
              done = 1;
           tecla = evento.xkey.keycode;
           break;

      }
    }

    if( tecla != 0 )
          LeerMovimiento ( tecla, &serpiente1 );

       GeneraComida( display, pixmap, window, gc );
       Movimiento( display, pixmap, window, gc, &serpiente1 );
       DibujaSerpiente(display, pixmap, window, gc, &serpiente1 );
       BorraSerpiente(display, pixmap, window, gc, &serpiente1 );

    /* coger el tiempo actual */
     gettimeofday(&end, NULL);

    /* calculamos el tiempo que ha pasado desde el anterior
       fotograma, para poder cronometrar el estado del juego */
    if( end.tv_usec < start.tv_usec )
        end.tv_usec += 1000000;
    time_dif = end.tv_usec - start.tv_usec;

    XSync(display, False);
    usleep(VELOCIDAD-time_dif);

    /* implementamos un cronometro de useg a segundos */
    tiempo += VELOCIDAD;
    if( tiempo > 1000000 )
    {
       tiempo -= 1000000;
       segundos++;
    }

 }

  sleep(1);
  printf("XSnake - Version 1.0 para X Window.\n\n");
  printf(" Has conseguido %d puntos.\n", serpiente1.puntos );
  printf(" Has sobrevivido %d minuto(s) y %d segundo(s).\n\n",
            segundos/60, segundos%60 );
  XUnmapWindow(display, window);
  XUnloadFont(display, fuente);
  XFreePixmap(display, pixmap);
  XCloseDisplay(display);
  exit( 0 );

}


/*---------------------------------------------------------
  Inicializa la matriz que se le pasa como parmetro
  poniendo a cero todos sus elementos.
 --------------------------------------------------------*/
void InicializaMatrices( void )
{
  int x, y;
  for( y=0; y<ymax+1; y++ )
    for( x=0; x<xmax+1; x++ )
       pantalla[x][y] = cola[x][y] = 0;
}


/*---------------------------------------------------------
  Inicializa los valores bsicos para la serpiente
  para preparar su posicionamiento y direccion inicial.
 --------------------------------------------------------*/
void InicializaSerpiente( struct serpiente *serp )
{
  serp->cabserx = xmax/2;
  serp->cabsery = ymax/2;
  serp->colaserx = xmax/2;
  serp->colasery = ymax/2;
  serp->espera = INCREMENTO/2;
  serp->izquierda = false;
  serp->derecha = true;
  serp->arriba = false;
  serp->abajo = false;
  serp->puntos = 0;

  pantalla[serp->colaserx][serp->colasery] = SERPIENTE;
  cola[serp->colaserx][serp->colasery] = DIR_DER;
}


/*---------------------------------------------------------
  Dibuja el tablero tanto en pantalla como en el pixmap
  que hace de pantalla virtual.
 --------------------------------------------------------*/
void DibujaTablero( Display *display, Pixmap pixmap,
                     Drawable window, GC gc )
{
  int x, y;

  for( x=0; x<=xmax; x++ )
  {
    DibujaBloque( display, pixmap, window, gc, x, ymin, PARED );
    DibujaBloque( display, pixmap, window, gc, x, ymax, PARED );
  }

  for( y=0; y<=ymax; y++ )
  {
    DibujaBloque( display, pixmap, window, gc, xmin, y, PARED );
    DibujaBloque( display, pixmap, window, gc, xmax, y, PARED );
  }

  XSetForeground(display, gc, verde);
  XDrawString( display, pixmap, gc, 10,
               ((ALTURA_BLOQUE*ymax)+40)-5, "PUNTOS: ", 8 );
  XDrawString( display, window, gc, 10,
               ((ALTURA_BLOQUE*ymax)+40)-5, "PUNTOS: ", 8 );
  XDrawString( display, pixmap, gc, ((ANCHURA_BLOQUE*(xmax-1))/2)-15,
               ((ALTURA_BLOQUE*ymax)+40)-5, "X-SNAKE", 7 );
  XDrawString( display, window, gc, ((ANCHURA_BLOQUE*(xmax-1))/2)-15,
               ((ALTURA_BLOQUE*ymax)+40)-10, "X-SNAKE", 7 );
}


/*---------------------------------------------------------
  Actualiza abajo los puntos que lleva el jugador.
 --------------------------------------------------------*/
void ActualizaPuntos( Display *display, Pixmap pixmap,
         Drawable window, GC gc, struct serpiente *serp )
{
  char puntos[7];

  sprintf(puntos, "%d", serp->puntos);

  XSetForeground(display, gc, negro);
  XFillRectangle( display, window, gc, 105,
                  ((ALTURA_BLOQUE*ymax)+40)-20, 6*9, 20 );
  XFillRectangle( display, pixmap, gc, 105,
                  ((ALTURA_BLOQUE*ymax)+40)-20, 6*9, 20 );
  XSetForeground(display, gc, amarillo);
  XDrawString( display, window, gc, 110,
                  ((ALTURA_BLOQUE*ymax)+40)-5, puntos, strlen(puntos) );
  XDrawString( display, pixmap, gc, 110,
                  ((ALTURA_BLOQUE*ymax)+40)-5, puntos, strlen(puntos) );
}

/*---------------------------------------------------------
  Realiza la accion apropiada segun la pulsacion de
  direccion realizada por el jugador.
 --------------------------------------------------------*/
void LeerMovimiento( char keycode, struct serpiente *serp )
{

   if( keycode == ARRIBA && serp->abajo == false )
   {
      serp->arriba = true;
      serp->derecha = serp->izquierda = serp->abajo = false;
   }

   else if( keycode == ABAJO && serp->arriba == false )
   {
      serp->abajo = true;
      serp->derecha = serp->arriba = serp->izquierda = false;
   }

   else if( keycode == IZQUIERDA && serp->derecha == false)
   {
      serp->izquierda = true;
      serp->derecha = serp->arriba = serp->abajo = false;
   }

   else if( keycode == DERECHA && serp->izquierda == false)
   {
      serp->derecha = true;
      serp->arriba = serp->izquierda =  serp->abajo = false;
   }
   tecla = 0;
}



/*---------------------------------------------------------
  Realiza la accion apropiada segun la pulsacion de
  direccion realizada por el jugador.
 --------------------------------------------------------*/
void Movimiento( Display *display, Pixmap pixmap,
                 Window window, GC gc, struct serpiente *serp )
{
   int dircola, cabserantx, cabseranty;
   char dato;

   cabserantx = serp->cabserx;
   cabseranty = serp->cabsery;

   if ( serp->izquierda == true )
   {
     if ( serp->cabserx > xmin+1 )
     {
        serp->cabserx = (serp->cabserx)-1;
        dircola = DIR_IZQ;
     }
     else done = 1;
   }

   else if ( serp->derecha == true )
   {
     if ( serp->cabserx < xmax-1 )
     {
        serp->cabserx = (serp->cabserx)+1;
        dircola = DIR_DER;
     }
     else done = 1;
   }

   else if ( serp->arriba == true )
   {
     if ( serp->cabsery > ymin+1 )
     {
        serp->cabsery = (serp->cabsery)-1;
        dircola = DIR_ARR;
     }
     else done = 1;
   }

   else if ( serp->abajo == true )
   {
     if ( serp->cabsery < ymax-1 )
     {
        serp->cabsery = (serp->cabsery)+1;
        dircola = DIR_ABJO;
     }
     else done = 1;
   }

   if( pantalla[serp->cabserx][serp->cabsery] == SERPIENTE )
        done = 1;

   else
   {

     if( pantalla[serp->cabserx][serp->cabsery] >= COMIDA1 )
     {
        BorraComida( display, pixmap, window, gc,
                      serp->cabserx, serp->cabsery );
        comida--;
        serp->espera = serp->espera + INCREMENTO;
        serp->puntos = serp->puntos + 10;
        ActualizaPuntos(display, pixmap, window, gc, serp );

     }


      pantalla[serp->cabserx][serp->cabsery] = SERPIENTE;
      cola[cabserantx][cabseranty] = dircola;
   }
}


/*---------------------------------------------------------
  Dibuja el grfico para la comida.
 --------------------------------------------------------*/
void DibujaComida( Display *display, Pixmap pixmap,
                    Window window, GC gc )
{
   int posx, posy, contador = 0;

   /* damos 100 intentos para encontrar un hueco donde
      posicionar la comida... (para cuando la pantalla
      ya est colapsada con nuestra serpiente) */
   do
   {
      posx = rand() % (xmax-6)+3;
      posy = rand() % (ymax-6)+3;
      contador++;
   }
   while( pantalla[posx][posy] != VACIO ||
          pantalla[posx+1][posy] != VACIO ||
          pantalla[posx][posy+1] != VACIO ||
          pantalla[posx+1][posy+1] != VACIO &&
          contador < 100 );
   if( contador < 100 )
   {
      comida++;

      pantalla[posx][posy] = COMIDA1;
      pantalla[posx+1][posy] = COMIDA2;
      pantalla[posx][posy+1] = COMIDA3;
      pantalla[posx+1][posy+1] = COMIDA4;

      XCopyArea(display, sprites[3], pixmap, gc,
           0, 0, ANCHURA_BLOQUE*2, ALTURA_BLOQUE*2, posx*ANCHURA_BLOQUE,
                                                  posy*ALTURA_BLOQUE);
     XCopyArea(display, sprites[3], window, gc,
           0, 0, ANCHURA_BLOQUE*2, ALTURA_BLOQUE*2, posx*ANCHURA_BLOQUE,
                                                  posy*ALTURA_BLOQUE);
  }
}


/*---------------------------------------------------------
  Borra grfico para la comida.
 --------------------------------------------------------*/
void BorraComida( Display *display, Pixmap pixmap,
                    Window window, GC gc, int x, int y )
{

  switch( pantalla[x][y] )
  {
     case COMIDA1:
       pantalla[x+1][y] = pantalla[x][y+1] = pantalla[x+1][y+1] = VACIO;
       DibujaBloque( display, pixmap, window, gc, x+1, y, VACIO );
       DibujaBloque( display, pixmap, window, gc, x, y+1, VACIO );
       DibujaBloque( display, pixmap, window, gc, x+1, y+1, VACIO );
       break;

     case COMIDA2:
       pantalla[x-1][y] = pantalla[x][y+1] = pantalla[x-1][y+1] = VACIO;
       DibujaBloque( display, pixmap, window, gc, x-1, y, VACIO );
       DibujaBloque( display, pixmap, window, gc, x, y+1, VACIO );
       DibujaBloque( display, pixmap, window, gc, x-1, y+1, VACIO );
       break;

     case COMIDA3:
       pantalla[x+1][y] = pantalla[x][y-1] = pantalla[x+1][y-1] = VACIO;
       DibujaBloque( display, pixmap, window, gc, x+1, y, VACIO );
       DibujaBloque( display, pixmap, window, gc, x, y-1, VACIO );
       DibujaBloque( display, pixmap, window, gc, x+1, y-1, VACIO );
       break;

     case COMIDA4:
       pantalla[x][y-1] = pantalla[x-1][y] = pantalla[x-1][y-1] = VACIO;
       DibujaBloque( display, pixmap, window, gc, x, y-1, VACIO );
       DibujaBloque( display, pixmap, window, gc, x-1, y, VACIO );
       DibujaBloque( display, pixmap, window, gc, x-1, y-1, VACIO );
       break;
  }
   pantalla[x][y] = SERPIENTE;
}


/*---------------------------------------------------------
  Genera la comida si es necesaria.
 --------------------------------------------------------*/
void GeneraComida( Display *display, Pixmap pixmap,
                    Window window, GC gc )
{
   if( comida < MAX_COMIDA )
      DibujaComida( display, pixmap, window, gc );

}


/*---------------------------------------------------------
  Dibuja el grfico para la serpiente.
 --------------------------------------------------------*/
void DibujaSerpiente( Display *display, Pixmap pixmap,
                       Window window, GC gc,
                       struct serpiente *serp )
{
   DibujaBloque( display, pixmap, window, gc,
                 serp->cabserx, serp->cabsery, SERPIENTE );
}


/*---------------------------------------------------------
  Elimina la cola si procede..
 --------------------------------------------------------*/
void EliminaCola( Display *display, Pixmap pixmap,
                   Window window, GC gc,
                   struct serpiente *serp )
{
 int direccion;

    direccion = cola[serp->colaserx][serp->colasery];
    DibujaBloque( display, pixmap, window, gc,
                 serp->colaserx, serp->colasery, VACIO );

    pantalla[serp->colaserx][serp->colasery] = VACIO;
    cola[serp->colaserx][serp->colasery] = VACIO;
    if( direccion == DIR_IZQ )
      serp->colaserx = (serp->colaserx)-1;
    if( direccion == DIR_DER )
      serp->colaserx = (serp->colaserx)+1;
    if( direccion == DIR_ARR )
      serp->colasery = (serp->colasery)-1;
    if( direccion == DIR_ABJO )
      serp->colasery = (serp->colasery)+1;
}


/*---------------------------------------------------------
  Borra la serpiente si procede.
 --------------------------------------------------------*/
void BorraSerpiente( Display *display, Pixmap pixmap,
                      Window window, GC gc,
                      struct serpiente *serp )
{

  if ( serp->espera == 0 )
     EliminaCola( display, pixmap, window, gc, serp );
  else
      serp->espera = serp->espera - 1;
}


/*--- Dibuja un bloque ----------------------------------*/
void DibujaBloque( Display *display, Drawable pixmap,
                   Drawable window, GC gc, int x, int y,
                   char tipo )
{
   DibujaBitmap( display, pixmap, gc, sprites[tipo],
                  x*ANCHURA_BLOQUE, y*ALTURA_BLOQUE);
   DibujaBitmap( display, window, gc, sprites[tipo],
                  x*ANCHURA_BLOQUE, y*ALTURA_BLOQUE);

}



/*--- Funcion que redibuja la pantalla necesaria --------*/
void RedibujaPantalla( Display *display, Drawable drawable,
   GC gc, Pixmap pixmap, int x, int y, int anchura, int altura  )
{
   XCopyArea(display, pixmap, drawable, gc,
             x, y, anchura, altura, x, y);
}


/*--- Funcion que dibuja un bitmap ----------------------*/
void DibujaBitmap( Display *display, Drawable drawable,
                   GC gc, Pixmap pixmap, int x, int y )
{
    XCopyArea(display, pixmap, drawable, gc,
              0, 0, ANCHURA_BLOQUE, ALTURA_BLOQUE, x, y);
}


/*--- Rutina para localizar colores por nombre ----------*/
unsigned long ColorPorNombre( Display *display, char *Nombre )
{
  XColor color, temp;
  XAllocNamedColor( display,
     DefaultColormap(display,DefaultScreen(display)),
     Nombre, &color, &temp );
  return( color.pixel );
}
